/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


#include <xacml_client.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* B: prototypes */



/* Helper funcs */

int pep_register_obligation_handlers (xacml_request_t);
int pep_construct_request_subject (xacml_request_t   request,
                           char *            subject_dn,
                           char *            issuer_dn,
                           int               subject_certificate_serial_number,
                           int               subject_ca_serial_number,
                           char *            subject_vo[],
                           int               subject_vo_cnt,
                           char *            voms_signing_subject[],
                           int               voms_signing_subject_cnt,
                           char *            voms_signing_issuer[],
                           int               voms_signing_issuer_cnt,
                           char *            voms_dns_port[],
                           int               voms_dns_port_cnt,
                           char *            subject_voms_fqan[],
                           int               subject_voms_fqan_cnt,
                           char *            subject_voms_primary_fqan,
                           char *            ca_policy_oid[],
                           int               ca_policy_oid_cnt,
                           char *            cert_chain
                          );
int pep_construct_request_resource (xacml_request_t   request,
                                    char *            resource_node_type,
                                    char *            host_dns_name,
                                    char *            resource_id,
                                    char *            resource_x509_id,
                                    char *            resource_x509_issuer
                                   );
int pep_construct_request_action   (xacml_request_t   request,
                                    char *            action_type
                                   );

int get_gidlist(
        const char * username,
        int * ngroups,
        gid_t ** group_list
    );

/* Obligation handlers */

int
default_handler(
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[]);


/* example */
int
local_user_name_handler(
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[]);
/* example */

int
Username_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[]);

int
UIDGID_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[]);

int
SecondaryGIDs_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[]);

int         
AFSToken_handler (
    void *                              handler_arg,
    const xacml_response_t              response,
    const char *                        obligation_id,
    xacml_effect_t                      fulfill_on,
    const char *                        attribute_ids[],
    const char *                        datatypes[],
    const char *                        values[]);
/* Obligation handlers */


int getOHUIDGIDFired(void);
int getOHSecondaryGIDsFired(void);
int getOHUsernameFired(void);

void setOHUIDGIDFired(int);
void setOHSecondaryGIDsFired(int);
void setOHUsernameFired(int);


/* E: prototypes */

