/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


#include "xacml_client.h"
#include "xacml_io_ssl.h"
#include "ssl-common.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

/* B: prototypes */

#include "common.h"
#include "pep_obligation_handlers.h"
#include "xacml_attr_obl_profile.h"

/* E: prototypes */





int main(int argc, char *argv[])
{
    xacml_request_t request;
    int ch;
    int i;
    int rc = 0;
    int trycnt = 5;
    char * ca_path = NULL;
    char * endpoint = NULL;
    /* char * use_io_module = NULL; */
    xacml_response_t response;

 /* 
    xacml_resource_attribute_t ra;
    const char *resattr[2];
  */

    char * fqans[] = { "/gin.ggf.org", "/gin.ggf.org/pragma" };
    int     fqan_cnt = 2;

    saml_status_code_t code;
    xacml_decision_t decision;


    debugmode = 0;
    xacml_init();
    xacml_request_init(&request);

    while ((ch = getopt(argc, argv, "e:c:k:a:dim:h")) != -1)
    {
        switch (ch)
        {
        case 'e':
            endpoint = optarg;
            break;
        case 'c':
            certFilePub = optarg;
            break;
        case 'k':
            certFilePriv = optarg;
            break;
        case 'a':
            ca_path = optarg;
            break;
        case 'm':
            xacml_request_set_io_module(request, optarg);
            break;
        case 'i':
            xacml_request_set_io_descriptor(request,
                        &xacml_io_ssl_descriptor);
            break;
        case 'd':
            debugmode = 1;
            break;
        case 'h':
        case '?':
        default:
            printf("Usage %s [-d] [-e endpoint] [-c certfile] [-k keyfile] [-a CA-path]\n",
                    argv[0]);
            exit(0);
        }
    }


    xacml_response_init(&response);


    /************ B: Subject *********
    xacml_request_set_subject(
            request,
            "/C=NL/CN=Oscar Koeroo");

    xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_X509_ID,
            XACML_DATATYPE_STRING,
            "",
            "/C=NL/CN=Oscar Koeroo");

    xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_SUBJECT_CA_SERIAL_NUMBER,
            XACML_DATATYPE_INTEGER,
            "",
            "2");

    xacml_request_add_subject_attribute(
            request,
            XACML_SUBJECT_CATEGORY_ACCESS_SUBJECT,
            XACML_ATTR_PROFILE_NS_URL_CA_POLICY_OID,
            XACML_DATATYPE_STRING,
            "",
            "010.120201.10101032.3042423.343232");
    ************ E: Subject *********/


    /************ B: Resource *********
    resattr[0] = "https://140.221.36.11:8081/wsrf/services/SecureCounterService";

    xacml_resource_attribute_init(&ra);

    xacml_resource_attribute_add(
            ra,
            XACML_RESOURCE_ATTRIBUTE_RESOURCE_ID,
            XACML_DATATYPE_STRING,
            "",
            resattr[0]);

    xacml_request_add_resource_attribute(request, ra);
    xacml_resource_attribute_destroy(ra);
    ************ E: Resource *********/

    /************ B: Action *********
    xacml_request_add_action_attribute(
            request,
            XACML_ACTION_ATTRIBUTE_ACTION_NAMESPACE,
            XACML_DATATYPE_STRING,
            "",
            "http://www.gridforum.org/namespaces/2003/06/ogsa-authorization/saml/action/operation");
    xacml_request_add_action_attribute(
            request,
            XACML_ACTION_ATTRIBUTE_ACTION_ID,
            XACML_DATATYPE_STRING,
            "",
            "createCounter");
    ************ E: Action *********/

    


    /************ B: Construction of the XACML request ***********/
    if (pep_construct_request_subject (request,
                                       "/C=NL/CN=Oscar Koeroo",    /* Subject DN */
                                       "/C=NL/CN=DutchGrid CA",    /* Issuer DN */
                                       123,                        /* Subject certificate serial number */
                                       1,                          /* Issuer certificate serial number */
                                       NULL,                       /* subject VO name (list of char *) */
                                       0,                          /* List (previous) member quantity */
                                       NULL,                       /* Subject DN of VOMS signing host */
                                       0,                          /* List (previous) member quantity */
                                       NULL,                       /* Issuer DN (CA) of VOMS signing host */
                                       0,                          /* List (previous) member quantity */
                                       NULL,                       /* DNS Port String f VOMS host */
                                       0,                          /* List (previous) member quantity */
                                       fqans,                      /* VOMS FQANs */
                                       fqan_cnt,                   /* List (previous) member quantity */
                                       "/gin.ggf.org",             /* Primary VOMS FQAN */
                                       NULL,                       /* CA Policy OIDs */
                                       0,                          /* List (previous) member quantity */
                                       NULL                        /* Certificate Chain in Base64 */
                                      ))
    {
        fprintf (stderr, "Error in constructing request subject\n");
        return 1;
    }

    if (pep_construct_request_resource (request,
                                        XACML_ATTR_PROFILE_NS_URL_RESOURCE_NODE_TYPE_CE,
                                        "localhost"
                                       ))
    {
        fprintf (stderr, "Error in constructing request resource\n");
        return 1;
    }
       
    if (pep_construct_request_action (request,
                                      XACML_ATTR_PROFILE_NS_URL_ACTION_RUN_TYPE_EXEC_NOW
                                     ))
    {
        fprintf (stderr, "Error in constructing request action\n");
        return 1;
    }
    /************ E: Construction of the XACML request ***********/
    


    /************ B: Register Obligation Handlers *********/
    if (pep_register_obligation_handlers (request))
    {
        fprintf (stderr, "Error in registration of obligation handlers\n");
        return 1;
    }
    /************ E: Register Obligation Handlers *********/


    if (certFilePub == NULL || certFilePriv == NULL)
    {
        fprintf (stderr, "Error: can't function without a certificate\n");
        abort();
    }

    if (endpoint == NULL)
    {
        endpoint = "http://localhost:8080/wsrf/services/XACML";
    }


    ch = xacml_query(endpoint,
                     request,
                     response);

    if (ch != 0)
    {
        for (i = 0; i < trycnt; i++)
        {
            printf("Error processing messages, try(%d)ing again.\n", i + 1);
            ch = xacml_query(endpoint,
                             request,
                             response);

            if (ch == 0) break;
        }
        if (ch != 0)
        {
            printf("Error processing messages\n");
            rc = 1;
            goto cleanup;
        }
    }

    xacml_response_get_saml_status_code(response, &code);
    xacml_response_get_xacml_decision(response, &decision);

    printf("Server said: %s:%d\n", saml_status_code_strings[code], decision);


cleanup:
    printf ("-> Can't do it: Destroying: obligation default_handler\n");
/*    xacml_obligation_destroy(default_handler); */
    printf ("-> Can't do it: Destroying: obligation localusername handler\n");
/*    xacml_obligation_destroy(local_user_name_handler); */
    printf ("Destroying: response\n");
    xacml_response_destroy(response);
    printf ("Destroying: request\n");
    xacml_request_destroy(request);


    return rc;
}

