/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */


#ifndef SSL_COMMON_H
 #define SSL_COMMON_H

#include <stdlib.h>
#include <limits.h>
#include <math.h>

#include <sys/types.h>
#include <pwd.h>

#include <unistd.h>

#include <string.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#include <openssl/ssl.h>
#include <openssl/x509v3.h>


#define CIPHER_LIST "ALL:!ADH:!LOW:!EXP:@STRENGTH"
#define MAXCHAINDEPTH 99


int init_OpenSSL(void);
void seed_prng(void);

int changeEffectiveToRealUid (uid_t *, uid_t *);
int restoreEffectiveToRealUid (uid_t, uid_t);


char * get_hostname_to_match_client_cert(void);
int    set_hostname_to_match_client_cert(char * override_h_n);
void  free_hostname_to_match_client_cert(void);

int SSL_recv (SSL * ssl, void * data, int num);

SSL * SSL_client_connect (int opensocket, SSL_CTX * ctx);
SSL * SSL_server_accept (int opensocket, char * remote_host, SSL_CTX * ctx);
int SSL_connect_shutdown (SSL * ssl);

SSL_CTX *setup_client_ctx(char * cafile, 
                          char * cadir, 
                          char * clientcert, 
                          char * clientkey);
SSL_CTX *setup_server_ctx(char * cafile,
                          char * cadir,
                          char * servercert,
                          char * serverkey);

int x509IsCA (X509 *cert);
char * x509_chain_to_dn(X509 * px509, STACK_OF(X509) *chain);
int print_ssl_error_msg (SSL * ssl, int ret, int printmsg, int * error_code);

long post_connection_check(SSL *ssl, const char *host);

extern SSL_CTX * ssl_ctx;
extern char * certCAPath;
extern char * certFilePub;
extern char * certFilePriv;
extern char * certOwner;

time_t my_timegm(struct tm *tm);

/************************/
#ifndef DISABLED_PROXY_SUPPORT
unsigned long grid_X509_knownCriticalExts(X509 *cert);
unsigned long grid_verifyProxy( STACK_OF(X509) *certstack );
time_t grid_asn1TimeToTimeT(char *asn1time, size_t len);
int scas_verify_callback(int ok, X509_STORE_CTX *store);
#if OPENSSL_VERSION_NUMBER < 0x00908000L
 int grid_X509_check_issued_wrapper(X509_STORE_CTX *ctx,X509 *x,X509 *issuer);
#endif
#endif /* DISABLED_PROXY_SUPPORT */

#if 0
int
scas_gsi_callback_X509_verify_cert(
    X509_STORE_CTX *                    context,
    void *                              arg);
#endif

#define GLOBUS_TRUE    1
#define GLOBUS_FALSE   0
#define GLOBUS_NULL    0
#define GLOBUS_SUCCESS 0
#define GLOBUS_FAILURE  -1


typedef enum globus_gsi_cert_utils_cert_type_e
{
    /** A end entity certificate */
    GLOBUS_GSI_CERT_UTILS_TYPE_EEC,
    /** A CA certificate */
    GLOBUS_GSI_CERT_UTILS_TYPE_CA,
    /** A X.509 Proxy Certificate Profile (pre-RFC) compliant
     *  impersonation proxy
     */
    GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_IMPERSONATION_PROXY,
    /** A X.509 Proxy Certificate Profile (pre-RFC) compliant
     *  independent proxy
     */
    GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_INDEPENDENT_PROXY,
    /** A X.509 Proxy Certificate Profile (pre-RFC) compliant
     *  limited proxy
     */
    GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_LIMITED_PROXY,
    /** A X.509 Proxy Certificate Profile (pre-RFC) compliant
     *  restricted proxy
     */
    GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_RESTRICTED_PROXY,
    /** A legacy Globus impersonation proxy */
    GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_PROXY,
    /** A legacy Globus limited impersonation proxy */
    GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_LIMITED_PROXY,
    /** A X.509 Proxy Certificate Profile RFC compliant impersonation proxy */
    GLOBUS_GSI_CERT_UTILS_TYPE_RFC_IMPERSONATION_PROXY,
    /** A X.509 Proxy Certificate Profile RFC compliant independent proxy */
    GLOBUS_GSI_CERT_UTILS_TYPE_RFC_INDEPENDENT_PROXY,
    /** A X.509 Proxy Certificate Profile RFC compliant limited proxy */
    GLOBUS_GSI_CERT_UTILS_TYPE_RFC_LIMITED_PROXY,
    /** A X.509 Proxy Certificate Profile RFC compliant restricted proxy */
    GLOBUS_GSI_CERT_UTILS_TYPE_RFC_RESTRICTED_PROXY
} globus_gsi_cert_utils_cert_type_t;

#define GSI_CERT_IS_PROXY(cert_type) \
        (cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_IMPERSONATION_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_INDEPENDENT_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_LIMITED_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_GSI_3_RESTRICTED_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_RFC_IMPERSONATION_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_RFC_INDEPENDENT_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_RFC_LIMITED_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_RFC_RESTRICTED_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_PROXY || \
         cert_type == GLOBUS_GSI_CERT_UTILS_TYPE_GSI_2_LIMITED_PROXY)


#endif /* COMMON_H */


