/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */



/*!
    \page lcmaps_scas_client.mod Site Central Authorization Service client plugin
 
    \section posixsynopsis SYNOPSIS
    \b lcmaps_scas_client.mod
 
    \section posixdesc DESCRIPTION
 
description (need to fill in)

 
*/


/*!
    \file   lcmaps_scas_client.c
    \brief  Plugin to contact the Site Central Authorization Service
    \author Oscar Koeroo for the EGEE.
*/





/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcmaps_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/x509_vfy.h>
#include <openssl/asn1.h>
#include <time.h>

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <lcmaps/lcmaps_modules.h>
#include <lcmaps/lcmaps_arguments.h>
#include <lcmaps/lcmaps_cred_data.h>
//    #include <lcmaps/lcmaps_voms_utils.h>
#include <lcmaps/lcmaps_vo_data.h>
#include <lcmaps/lcmaps_types.h>

//    #include <voms/voms_apic.h>


//#undef TESTBIO
//#define TESTBIO
#ifdef TESTBIO
#    include <openssl/pem.h>
#    include <openssl/bio.h>
#endif


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "parameters.h"


/*** SAML2-XACML2 ***/
#include "xacml_client.h"
#include "xacml_io_ssl.h"
#include "common.h"
#include "ssl-common.h"

#include "pep_obligation_handlers.h"
#include "xacml_attr_obl_profile.h"

/*** SAML2-XACML2 ***/


/******************************************************************************
                                Definitions
******************************************************************************/

#define PLUGIN_RUN      0
#define PLUGIN_VERIFY   1


// #define VOMS_BUFFER_SIZE  1024

// What do we want to do with this time presentation?
#define GMT               1
#define LOCAL             2
#define LEAVE_TIME_ITS_OK 3

typedef enum ENDPOINT_SELECTION_E {
    ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM = 1,
    ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START  = 2,
    ENDPOINT_SELECTION_RANDOM                    = 3 
} ENDPOINT_SELECTION_T;
 

typedef enum XACML_QUERY_ANSWER_E {
    XACML_QUERY_ALLOWED          = 0,
    XACML_QUERY_BANNED           = 1,
    XACML_QUERY_PROTOCOL_ERROR   = 2 
} XACML_QUERY_ANSWER_T;

/******************************************************************************
                          Module specific prototypes
******************************************************************************/


int plugin_initialize(int, char **);
int plugin_introspect(int *, lcmaps_argument_t **);
int plugin_run(int, lcmaps_argument_t *);
int plugin_verify(int, lcmaps_argument_t *);
int plugin_terminate(void);
static int plugin_run_or_verify(int, lcmaps_argument_t *, int);

/*
int x509IsCA (X509 *cert);
*/
int collapseFqansFromVOMSData (lcmaps_vomsdata_t *   vomsdata,
                               char              *** fqans,
                               int               *   nfqans);
int pem_string_to_x509(X509 **px, char *certstring);
static XACML_QUERY_ANSWER_T xacmlqueryscas (char *           , 
                                            xacml_request_t  , 
                                            xacml_response_t );


/******************************************************************************
                       Define module specific variables
******************************************************************************/

#define MAX_ENDPOINTS         32
#define MAX_ENDPOINT_BUFSIZE  256

/* 
static int    endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM;
static int    endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
*/
static int    endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
static char   endpoints[MAX_ENDPOINTS][MAX_ENDPOINT_BUFSIZE];
static int    n_endpoints = 0;
static int    endpoint_retrycount = 2; /* By default each endpoint is tried twice in a row for a complete protocol stack retry. */
static char * resourcetype = NULL;
static char * actiontype = NULL;
/*
char * certFilePub  = NULL;
char * certFilePriv = NULL;
char * ca_path      = NULL;
*/

static int setResourceX509ID = 0;
static int use_authorization_only = 0;
static int use_pool_index_bug_fix = 0;

/*****************************************************************************/




/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[0]: the name of the plugin
Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
    LCMAPS_MOD_NOFILE  : db file not found (will halt LCMAPS initialization)
******************************************************************************/
int plugin_initialize(
        int argc,
        char ** argv
)
{
    char * logstr = "\tlcmaps_plugin_scas_client-plugin_initialize()";
    int i;

    certFilePub  = NULL;
    certFilePriv = NULL;
    certCAPath   = NULL;


    // connStr = (char *) malloc (sizeof(char) * 10000);
    // for (i = 0; i < 1000; i++) connStr[i] = '\0';
 
    lcmaps_log_debug(1,"%s: passed arguments:\n", logstr);
    for (i=0; i < argc; i++)
    {
       lcmaps_log_debug(2,"%s: arg %d is %s\n", logstr, i, argv[i]);
    }
 
    /*
     * CERTDIR = The directory which contains the CA certificates
     * VOMSDIR = The directory which contains the certificates of the VOMS servers 
     */
 
    /*
     * Parse arguments, argv[0] = name of plugin, so start with i = 1
     */

    for (i = 1; i < argc; i++)
    {
        /* Endpoint selection strategy */
        if ( ((strncasecmp(argv[i], "--endpoint-strategy", strlen("--endpoint-strategy")) == 0) || 
              (strncasecmp(argv[i], "-endpoint-strategy", strlen("-endpoint-strategy")) == 0)) && 
             (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if      (strcasecmp (argv[i + 1], "round-robin") == 0)              endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM;
                else if (strcasecmp (argv[i + 1], "round-robin-random-start") == 0) endpoint_selection_strategy = ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START;
                else if (strcasecmp (argv[i + 1], "random") == 0)                   endpoint_selection_strategy = ENDPOINT_SELECTION_RANDOM;
                else
                {
                    lcmaps_log (1, "%s: Error: --endpoint-strategy not understood. Options are: \"round-robin\", \"round-robin-random-start\" and \"random\".\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Endpoint */
        else if ( ( (strncasecmp(argv[i], "-endpoint", strlen("-endpoint")) == 0) || 
                    (strncasecmp(argv[i], "--endpoint", strlen("--endpoint")) == 0) ) && 
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                if (n_endpoints < MAX_ENDPOINTS)
                {
                    bzero (endpoints[n_endpoints], MAX_ENDPOINT_BUFSIZE);
                    if (strlen (argv[i + 1]) >= MAX_ENDPOINT_BUFSIZE)
                    {
                        lcmaps_log (1, "%s: Error: endpoint URL is too long: %s...\n", strncpy(endpoints[n_endpoints], argv[i + 1], MAX_ENDPOINT_BUFSIZE - 1), logstr);
                        return LCMAPS_MOD_FAIL;
                    }
                    else
                    {
                        strncpy(endpoints[n_endpoints], argv[i + 1], MAX_ENDPOINT_BUFSIZE - 1);
                        n_endpoints++;
                    }
                }
                else
                {
                    lcmaps_log (1, "%s: Error: Too many endpoints configured, the maximum amount of configurable endpoints is %d\n", logstr, MAX_ENDPOINTS);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Change the privileges of the process to this user account to read a proxy or certificate files from the file system */
        else if ( ((strncasecmp(argv[i], "-cert-owner",  strlen("-cert-owner"))  == 0) ||
                   (strncasecmp(argv[i], "--cert-owner", strlen("--cert-owner")) == 0)) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                certOwner = strdup(argv[i + 1]);
            }
            i++;
        }

        /* Endpoint retry count */
        else if ( ( (strncasecmp(argv[i], "-retry", strlen("-retry")) == 0) || 
                    (strncasecmp(argv[i], "--retry", strlen("--retry")) == 0) ) && 
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                endpoint_retrycount = strtol (argv[i + 1], NULL, 10);
                if (errno == EINVAL)
                {
                    lcmaps_log (1, "%s: Error: Couldn't parse the value \"%s\" to a number for the --retry option.\n", logstr, argv[i + 1]);
                    return LCMAPS_MOD_FAIL;
                }
                if (!(endpoint_retrycount > 0))
                {
                    lcmaps_log (1, "%s: Error: Endpoint retry count to low. It must be at least '1'\n", logstr);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }


        /* (optional) Host certificate file */
        else if ((((strlen(argv[i]) == 5) && (strncasecmp(argv[i], "-cert", 5) == 0)) ||
                  ((strlen(argv[i]) == 6) && (strncasecmp(argv[i], "--cert", 6) == 0)))
                 && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                certFilePub = strdup(argv[i + 1]);
            }
            i++;
        }

        /* (optional) Host certificate private key file */
        else if ( ((strncasecmp(argv[i], "-key", 4) == 0) ||
                   (strncasecmp(argv[i], "--key", 5) == 0)) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                certFilePriv = strdup(argv[i + 1]);
            }
            i++;
        }

        /* CA Path - path to CA certificate files (and CRLs) */
        else if ( ((strncasecmp(argv[i], "-capath", 7) == 0) ||
                   (strncasecmp(argv[i], "--capath", 8) == 0)) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                certCAPath = strdup(argv[i + 1]);
            }
            i++;
        }

        /* Resource Type - Set the resource type to CE, WN, RB, SE */
        else if ( ((strncasecmp(argv[i], "-resourcetype", 13) == 0) ||
                   (strncasecmp(argv[i], "--resourcetype", 14) == 0)) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "se") == 0)      resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_SE;
                else if (strcasecmp (tmprt, "ce") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_CE;
                else if (strcasecmp (tmprt, "wn") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_WN;
                else if (strcasecmp (tmprt, "rb") == 0) resourcetype = XACML_ATTR_PROFILE_NS_URL_RESOURCE_RESOURCE_TYPE_RB;
                else 
                {
                    lcmaps_log (1, "%s: Error: Resource type unknown: \"%s\". Only accepting: \"ce\", \"se\", \"rb\" and \"wn\"\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        /* Action Type - Set the action type to CE, WN, RB, SE */
        else if ( ((strncasecmp(argv[i], "-actiontype", 11) == 0) ||
                   (strncasecmp(argv[i], "--actiontype", 12) == 0)) && (i + 1 < argc))
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                char * tmprt = argv[i + 1];

                if (strcasecmp (tmprt, "access-file") == 0)      actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_ACCESS_FILE;
                else if (strcasecmp (tmprt, "queue") == 0)       actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_QUEUE;
                else if (strcasecmp (tmprt, "execute-now") == 0) actiontype = XACML_ATTR_PROFILE_NS_URL_ACTION_ACTION_TYPE_EXEC_NOW;
                else 
                {
                    lcmaps_log (1, "%s: Error: Action type unknown: \"%s\". Only accepting: \"queue\", \"execute-now\" and \"access-file\"\n", logstr, tmprt);
                    return LCMAPS_MOD_FAIL;
                }
            }
            i++;
        }

        else if ( (strncasecmp(argv[i], "-authorization-only",  strlen("-authorization-only" )) == 0) ||
                  (strncasecmp(argv[i], "--authorization-only", strlen("--authorization-only")) == 0) )
        {
            use_authorization_only = 1;
        }

        else if ( (strncasecmp(argv[i], "-enable-poolindex-fix",  strlen("-enable-poolindex-fix" )) == 0) ||
                  (strncasecmp(argv[i], "--enable-poolindex-fix", strlen("--enable-poolindex-fix")) == 0) )
        {
            use_pool_index_bug_fix = 1;
        }


        /* Override hostname - When evaluating the host certificate, override the expected Hostname to the following setting */
        else if ( ( (strncasecmp(argv[i], "-override-expected-hostname", strlen("-override-expected-hostname")) == 0) ||
                    (strncasecmp(argv[i], "--override-expected-hostname", strlen("--override-expected-hostname")) == 0) ) &&
                  (i + 1 < argc) )
        {
            if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
            {
                /* Clean the mem */
                free_hostname_to_match_client_cert();

                /* Set the expected hostname */
                set_hostname_to_match_client_cert(argv[i + 1]);
                lcmaps_log_debug (5, "%s: Setting expected hostname of the service to \"%s\"", logstr, argv[i + 1]);
            }
            i++;
        }

        /* Parameter unknown */
        else
        {
            lcmaps_log(1, "%s: Error in initialization parameter: %s (failure)\n", logstr, argv[i]);
            return LCMAPS_MOD_FAIL;
        }
    }

    /*
     * TODO: Add file open check on both -cert and -key set file paths (existance and such)
     */


    /* Without an endpoint, this plugin is pointless */
    if (n_endpoints == 0)
    {
        lcmaps_log (1, "%s: There are no endpoints configured to contact. Please configure at least one --endpoint <url> option.\n", logstr);
        return LCMAPS_MOD_FAIL;
    }


    for (i = 0; i < n_endpoints; i++)
    {
        lcmaps_log_debug (5, "%s: Using endpoint: %s\n", logstr, endpoints[i]);
    }
    switch (endpoint_selection_strategy)
    {
        case ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM :
            lcmaps_log_debug (5, "%s: Endpoint selection strategy set to \"round-robin, starting from the top of the list\"\n", logstr);
            break;
        case ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START  :
            lcmaps_log_debug (5, "%s: Endpoint selection strategy set to \"round-robin with random starting point in the list\"\n", logstr);
            break;
        case ENDPOINT_SELECTION_RANDOM                    :
            lcmaps_log_debug (5, "%s: Endpoint selection strategy set to \"random endpoint selection\"\n", logstr);
            break;
    }


    /*
     *  If the private key and/or the public key (certificate) is not set, 
     *  then try to look at the X509_USER_CERT and X509_USER_KEY environment vars for it
     */
    if (!certFilePub)
    {
        if (getenv ("X509_USER_CERT"))
        {
            certFilePub  = strdup(getenv ("X509_USER_CERT"));
            lcmaps_log_debug(3, "%s: Using as client SSL credential the certificate from $X509_USER_CERT \"%s\".\n", logstr, certFilePub);
        }
    }
    if (!certFilePriv)
    {
        if (getenv ("X509_USER_KEY"))
        {
            certFilePriv  = strdup(getenv ("X509_USER_KEY"));
            lcmaps_log_debug(3, "%s: Using as client SSL credential the private key from $X509_USER_KEY \"%s\".\n", logstr, certFilePriv);
        }
    }


    /*
     * Test if both the certificate file and private key file are set or none to go look for the $X509_USER_PROXY setting
     */
    if ((!certFilePub) && (certFilePriv)) 
    {
        lcmaps_log (1, "%s: Error: Only -key was set, not -cert. Both or neither must be set. Aborting to avoid confusion and/or unintentional use of the credentials\n", logstr);
        return LCMAPS_MOD_FAIL;
    }

    if ((certFilePub) && (!certFilePriv)) 
    {
        lcmaps_log (1, "%s: Error: Only -cert was set, not -key. Both or neither must be set. Aborting to avoid confusion and/or unintentional use of the credentials\n", logstr);
        return LCMAPS_MOD_FAIL;
    }


    /* The host credentials should be available here. */
    if (certFilePub && certFilePriv)
    {
        setResourceX509ID = 1; 
    }
 


    /*
     *  If the private key and/or the public key (certificate) is not set, 
     *  then try the X509_USER_PROXY environment variable 
     */
    /*
     *  if ((!certFilePub) && (!certFilePriv))
     *  {
     *      x509_user_proxy = getenv ("X509_USER_PROXY");
     *      certFilePub  = strdup(x509_user_proxy);
     *      certFilePriv = strdup(x509_user_proxy);
     *  }
     */

    /* 
     * If the private key and/or the certificate can not be found, then this is a clear initialization failure, unless the $X509_USER_PROXY provides for more information. 
     * This will come in the next stage, the plug-in run
     * The SSL connection to the service requires client side credentials 
     */
    if ((!certFilePub) || (!certFilePriv))
    {
        lcmaps_log_debug (3, "%s: Warning: No client SSL credentials provided yet. Will look for them at the $X509_USER_PROXY in the plug-in run stage\n", logstr);
        lcmaps_log_debug (5, "%s:          Checked at \'-cert\', \'-key\', $X509_USER_CERT and $X509_USER_KEY.\n", logstr);
        /* return LCMAPS_MOD_FAIL; */
    }
    else
    {
        lcmaps_log_debug(5, "%s: Using the following credentials to contact the service:\n", logstr);
        lcmaps_log_debug(5, "%s:       certificate: \"%s\"\n", logstr, certFilePub);
        lcmaps_log_debug(5, "%s:       private key: \"%s\"\n", logstr, certFilePriv);
    }


    if (!certCAPath)
    {
        lcmaps_log_debug (2, "%s: CA Path not set, using Default Certificate CA Path: %s\n", logstr, certCAPath);
        certCAPath = strdup("/etc/grid-security/certificates/");
    }

    lcmaps_log_debug(2,"%s: Initialization succeeded\n", logstr);
    return LCMAPS_MOD_SUCCESS;
} 





/******************************************************************************
Function:   plugin_introspect
Description:
    return list of required arguments
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_introspect(
        int * argc,
        lcmaps_argument_t ** argv
)
{
    char * logstr = "\tlcmaps_plugin_voms-plugin_introspect()";
    static lcmaps_argument_t argList[] = {
        { "user_dn"        , "char *"                ,  1, NULL},
        { "px509_cred"     , "X509 *"                ,  0, NULL},
        { "px509_chain"    , "STACK_OF(X509) *"      ,  0, NULL},
        { "voms_data_list" ,  "lcmaps_vomsdata_t *"  ,  0, NULL},
        { "nvoms_data"     ,  "int"                  ,  0, NULL},
        { "job_request"    , "char *"                ,  0, NULL},
        { NULL             , NULL                    , -1, NULL}
    };

    lcmaps_log_debug(3,"%s: introspecting\n", logstr);

    *argv = argList;
    lcmaps_log_debug(3,"%s: before lcmaps_cntArgs()\n", logstr);
    *argc = lcmaps_cntArgs(argList);
    lcmaps_log_debug(3,"%s: address first argument: 0x%x\n", logstr,argList);

    lcmaps_log_debug(2,"%s: Introspect succeeded\n", logstr);
    return LCMAPS_MOD_SUCCESS;
}


/******************************************************************************
Function:   plugin_run
Description:
    Gather credentials for LCMAPS
Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_run(
        int argc,
        lcmaps_argument_t * argv
)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_RUN);
}


/******************************************************************************
Function:   plugin_verify
Description:
    Verify if user is entitled to use local credentials based on his grid
    credentials. This means that the site should already have been set up
    by, e.g., LCMAPS in a previous run. This method will not try to setup
    account leases, modify (distributed) passwd/group files, etc. etc.
    The outcome should be identical to that of plugin_run().
    In this particular case "plugin_verify()" is identical to "plugin_run()"

Parameters:
    argc: number of arguments
    argv: list of arguments
Returns:
    LCMAPS_MOD_SUCCESS: authorization succeeded
    LCMAPS_MOD_FAIL   : authorization failed
******************************************************************************/
int plugin_verify(
        int argc,
        lcmaps_argument_t * argv
)
{
    return plugin_run_or_verify(argc, argv, PLUGIN_VERIFY);
}

static int plugin_run_or_verify(
        int argc,
        lcmaps_argument_t * argv,
        int lcmaps_mode
)
{
    char             *  local_host = NULL;
    char             *  resource_x509_id     = NULL;
    char             *  resource_x509_issuer = NULL;
    int                 cnt_dn = 0;


/*********** Beg: Initialization *********/    
    char             *  logstr      = "\tlcmaps_plugin_scas_client_run()";
    char             *  encoded_dn = "(void)";

    char             *  datetime = NULL;
    time_t              my_clock;
    struct tm        *  tmpTime;

    STACK_OF(X509)   *  myChain = NULL;
    int                 depth = 0;
/*    int                 certCount = 0; */

    int                 amount_of_CAs = 0;
    char             *  user_dn = NULL;
    char             *  issuer_dn = NULL;
    lcmaps_vomsdata_t * vomsdata  = NULL;
/*
    int                 nvomsdata = 0;
*/
    /* Only one ... for now */
    char             ** vonames         = malloc (1 * sizeof (char *));
    char             ** voms_issuer_dns = malloc (1 * sizeof (char *));
    char             ** voms_issuer_cas = malloc (1 * sizeof (char *));
    char             ** voms_uris       = malloc (1 * sizeof (char *));
          
    int                 i = 0;
    unsigned int        seed;
    int                 random_choice = 0;


/*********** Beg: SAML2-XACML2 *********/
    xacml_request_t request=NULL; /* NOTE: is pointer, make sure it's initialized */
    int rc = 0;
    int should_do_high_level_retry = 0;

/* Hack: SAML2-XACML2 must not write to stdout */
    int streams_redirected  = 0;
    int fd_out              = -1;
    int fd_out_save         = -1;
    int fd_err              = -1;
    int fd_err_save         = -1;
/* Hack: SAML2-XACML2 must not write to stdout */


#if PERFORMRETRY
    int trycnt = 3;
#endif
    xacml_response_t response=NULL; /* NOTE: is pointer, make sure it's initialized */

    char ** fqans  = NULL;
    int     nfqans = 0;

    debugmode = 0;
/*********** End: SAML2-XACML2 *********/
/*********** End: Initialization *********/    




    /*
     * The beginning
     */
    if (lcmaps_mode == PLUGIN_RUN)
        logstr = "\tlcmaps_plugin_scas_client-plugin_run()";
    else if (lcmaps_mode == PLUGIN_VERIFY)
        logstr = "\tlcmaps_plugin_scas_client-plugin_verify()";
    else
    {
        lcmaps_log(1,
            "\tlcmaps_plugin_scas_client-plugin_run_or_verify(): attempt to run plugin in invalid mode: %d\n",
            lcmaps_mode);
        goto fail_scas_client;
    }

 

    /*
     * In the beginning... there was time
     */


    time(&my_clock);
    tmpTime = gmtime(&my_clock);
    datetime = malloc(sizeof(char) * 20);
    snprintf(datetime, 20, "%04d-%02d-%02d %02d:%02d:%02d",
           tmpTime->tm_year + 1900, tmpTime->tm_mon + 1, tmpTime->tm_mday,
           tmpTime->tm_hour, tmpTime->tm_min, tmpTime->tm_sec);




    /*
     *  If the private key and the public key (certificate) are not set, 
     *  then try the X509_USER_PROXY environment variable 
     */
    if ((!certFilePub) && (!certFilePriv))
    {
        char * x509_user_proxy = NULL;

        if (!(x509_user_proxy = getenv ("X509_USER_PROXY")))
        {
            lcmaps_log(1, "%s: No client side credentials for SSL handshake to the SCAS service presented. Not in the configuration file and not in the $X509_USER_{PROXY,CERT,KEY} environment variables.\n", 
                          logstr);
            goto fail_scas_client;
        }
        else
        {
            lcmaps_log_debug (3, "%s: Using $X509_USER_PROXY \"%s\"\n", logstr, x509_user_proxy);
            
            certFilePub  = strdup(x509_user_proxy);
            certFilePriv = strdup(x509_user_proxy);
        }
    }



    /* Random function seeding */
    seed = (unsigned int)time((time_t *)NULL);
    srand(seed);
    /* Random function seeding */
    
    /***  Acquiring  information  ***/

    lcmaps_log_debug (2, "%s: Acquiring information from LCMAPS framework\n", logstr);
    if ( ( lcmaps_getArgValue("user_dn", "char *", argc, argv)) )
    {
        lcmaps_log_debug(5, "%s: found user_dn placeholder.\n", logstr);

        if ( !(user_dn  = *(char **) lcmaps_getArgValue("user_dn", "char *", argc, argv) ) )
        {
            lcmaps_log_debug (1, "%s: value of user_dn is empty. No user DN found by the framework in the proxy chain.\n", logstr);
            /* goto fail_scas_client; */
        }
        else
        {
            lcmaps_log_debug (5, "%s: user_dn = %s\n", logstr, user_dn);
        }
    }
    else
    {
        lcmaps_log_debug (2, "%s: could not get value of user dn from the framework!\n", logstr);
    }


    //////////////////////////////////
    //  Gimme my Bling Bling Chain  //
    //////////////////////////////////
    if ( (myChain = *(STACK_OF(X509) **) lcmaps_getArgValue("px509_chain", "STACK_OF(X509) *", argc, argv)) )
    {
        lcmaps_log_debug(5, "%s: found X.509 chain.\n", logstr);
    }
    else
    {
        lcmaps_log(1, "%s: could not get value of X.509 chain.\n", logstr);
        goto fail_scas_client;
    }


    depth = sk_X509_num(myChain);
    // How many CA certs are there in the chain?
    for (i = 0; i < depth; i++)
    {
        X509 * my_c = sk_X509_value(myChain, i);

        user_dn = X509_NAME_oneline(X509_get_subject_name(my_c), NULL, 0);
        lcmaps_log_debug (5,"%s: level[%d]: User DN is   : %s\n", logstr, i, user_dn);
        free(user_dn); user_dn=NULL;

        issuer_dn = X509_NAME_oneline(X509_get_issuer_name(my_c), NULL, 0);
        lcmaps_log_debug (5,"%s: level[%d]: Issuer DN is : %s\n", logstr, i, issuer_dn);
        free(issuer_dn); issuer_dn=NULL;

        if (x509IsCA(my_c))
            amount_of_CAs++;
    }
    lcmaps_log_debug (5, "%s: The number of certificates on the stack is: %d\n", logstr, depth);
    lcmaps_log_debug (5, "%s: The amount of CA certificates is: %d\n", logstr, amount_of_CAs);

    user_dn   = X509_NAME_oneline(X509_get_subject_name(sk_X509_value(myChain, depth - 1 - amount_of_CAs)), NULL, 0);
    lcmaps_log_debug (5,"%s: User DN is : %s\n", logstr, user_dn);

    issuer_dn = X509_NAME_oneline(X509_get_issuer_name(sk_X509_value(myChain, depth - 1 - amount_of_CAs)), NULL, 0);
    lcmaps_log_debug (5,"%s: Issuer DN is : %s\n", logstr, issuer_dn);

    
    /* Register DN in LCMAPS when found and not yet set */
    getCredentialData (DN, &cnt_dn);
    if (cnt_dn == 0)
    {
        /* Fix - Brian: Nov 12 2010 */
        /* addCredentialData creates an internal copy of user_dn */
        addCredentialData (DN, &user_dn);
    }

/********************************* New VOMS Structure *********************************/


    lcmaps_log_debug (5,"%s: Grabbing VOMS data\n", logstr, depth);
    vomsdata = NULL;

    if ( ( lcmaps_getArgValue("voms_data_list", "lcmaps_vomsdata_t *", argc, argv)) )
    {
        lcmaps_log_debug(5,"%s: found lcmaps voms_data_list placeholder.\n", logstr);

        if ( !(vomsdata = *(lcmaps_vomsdata_t **) lcmaps_getArgValue("voms_data_list", "lcmaps_vomsdata_t *", argc, argv)) )
        {
            lcmaps_log_debug (1,"%s: value of lcmaps_voms_data_list is empty. No VOMS AC(s) found by the framework in the proxy chain.\n", logstr);
            /* goto fail_scas_client; */
        }
        else
        {
            lcmaps_log_debug (5, "%s: vomsdata->nvoms = %d\n", logstr, vomsdata->nvoms);
            lcmaps_log_debug (5, "%s: vomsdata->voms = %x\n", logstr, vomsdata->voms);
            lcmaps_log_debug (5, "%s: vomsdata->voms[0] = %x\n", logstr, vomsdata->voms[0]);
            lcmaps_log_debug (5, "%s: vomsdata->voms[0].userdn = %s\n", logstr, vomsdata->voms[0].user_dn);
        }
    }
    else
    {
        lcmaps_log(0,"%s: No value of lcmaps_voms_data_list.\n", logstr);
    }


    /*************** 
       TODO: Losts of items are single valued now and must be changed to accomidate for a list of such values. !
             the list oms issuer dn and voms VO name are only send for the first VO in the list !
    ***************/
    /* This will collapse the FQANs into a simple array of strings */
    vonames[0]         = NULL;
    voms_issuer_dns[0] = NULL;
    voms_issuer_cas[0] = NULL;
    voms_uris[0]       = NULL;


    if (vomsdata && vomsdata->nvoms > 0)
    {
        collapseFqansFromVOMSData (vomsdata, &fqans, &nfqans);

        if (vomsdata->voms[0].voname)
            vonames[0] = strdup(vomsdata->voms[0].voname);

        if (vomsdata->voms[0].voms_issuer_dn)
            voms_issuer_dns[0] = strdup(vomsdata->voms[0].voms_issuer_dn);

        if (vomsdata->voms[0].voms_issuer_ca)
            voms_issuer_cas[0] = strdup(vomsdata->voms[0].voms_issuer_ca);

        if (vomsdata->voms[0].uri)
            voms_uris[0] = strdup(vomsdata->voms[0].uri);

        for (i = 0; i < nfqans; i++)
        {
            lcmaps_log_debug (5, "%s: Sending FQAN list[%d]: %s\n", logstr, i, fqans[i]);
        }
    }

/********************************* New VOMS Structure *********************************/



    /* Work goes here */
    lcmaps_log_debug (5, "%s: Starting XACML framework\n", logstr);

/*
    certFilePub  = "/etc/grid-security/hostcert.pem";
    certFilePriv = "/etc/grid-security/hostkey.pem";
    ca_path      = "/etc/grid-security/certificates/";
*/

    xacml_init();
    xacml_request_init(&request);

    xacml_request_set_io_descriptor(request, &xacml_io_ssl_descriptor);
    xacml_response_init(&response);

    lcmaps_log_debug (5, "%s: XACML framework initialized, grabbing information\n", logstr);

    /************ B: Construction of the XACML request ***********/


    /*************** 
       TODO: Losts of items are single valued now and must be changed to accomidate for a list of such values. !
             the list oms issuer dn and voms VO name are only send for the first VO in the list !
    ***************/
             
    if (pep_construct_request_subject (request,
                                       user_dn,                               /* Subject DN */
                                       issuer_dn,                             /* Issuer DN */
                                       0,                                     /* Subject certificate serial number */
                                       0,                                     /* Issuer certificate serial number */
                                       (vomsdata && vomsdata->nvoms > 0 && vomsdata->voms[0].voname) ? vonames : NULL,
                                                                              /* subject VO name (list of char *) */
                                       (vomsdata && vomsdata->nvoms > 0 && vomsdata->voms[0].voname) ? 1 : 0, 
                                                                              /* List (previous) member quantity */
                                       (vomsdata && vomsdata->nvoms > 0 && vomsdata->voms[0].voms_issuer_dn) ? voms_issuer_dns : NULL,
                                                                              /* Subject DN of VOMS signing host */
                                       (vomsdata && vomsdata->nvoms > 0 && vomsdata->voms[0].voms_issuer_dn) ? 1 : 0,
                                                                              /* List (previous) member quantity */
                                       (vomsdata && vomsdata->nvoms > 0 && vomsdata->voms[0].voms_issuer_ca) ? voms_issuer_cas : NULL,
                                                                              /* Issuer DN (CA) of VOMS signing host */
                                       (vomsdata && vomsdata->nvoms > 0 && vomsdata->voms[0].voms_issuer_ca) ? 1 : 0,
                                                                              /* List (previous) member quantity */
                                       (vomsdata && vomsdata->nvoms > 0 && vomsdata->voms[0].uri) ? voms_uris : NULL,
                                                                              /* DNS Port String of VOMS host */
                                       (vomsdata && vomsdata->nvoms > 0 && vomsdata->voms[0].uri) ? 1 : 0,
                                                                              /* List (previous) member quantity */
                                       fqans,                                 /* VOMS FQANs */
                                       nfqans,                                /* List (previous) member quantity */
                                       (nfqans > 0) ? fqans[0] : NULL,        /* Primary VOMS FQAN */
                                       NULL,                                  /* CA Policy OIDs */
                                       0,                                     /* List (previous) member quantity */
                                       NULL                                   /* Certificate Chain in Base64 */
                                      ))
    {
        lcmaps_log (1, "%s: XACML: Error in constructing request subject\n", logstr); 
        goto fail_scas_client;
    }
    /* Clean up */
    if (vonames)
    {
        if (vonames[0]) free(vonames[0]);
        free(vonames); vonames=NULL;
    }
    if (voms_issuer_dns)
    {
        if (voms_issuer_dns[0]) free(voms_issuer_dns[0]);
        free(voms_issuer_dns); voms_issuer_dns=NULL;
    }
    if (voms_issuer_cas)
    {
        if (voms_issuer_cas[0]) free(voms_issuer_cas[0]);
        free(voms_issuer_cas); voms_issuer_cas=NULL;
    }
    if (voms_uris)
    {
        if (voms_uris[0]) free(voms_uris[0]);
        free(voms_uris); voms_uris=NULL;
    }
    /* Clean up */


    /* Construct action */
    if (pep_construct_request_action (request,
                                      actiontype
                                     ))
    {
        lcmaps_log (1, "%s: XACML: Error in constructing request action\n", logstr); 
        goto fail_scas_client;
    }




    lcmaps_log_debug (5, "%s: Fetching DNS/Host name\n", logstr);
#ifndef MAXHOSTNAMELEN
   #define MAXHOSTNAMELEN 255
#endif /* MAXHOSTNAMELEN */
    local_host = (char *) malloc (sizeof(char) * MAXHOSTNAMELEN);
    gethostname (local_host, MAXHOSTNAMELEN);
    lcmaps_log_debug (5, "%s: DNS/Host name is: %s\n", logstr, local_host);


    /* Add resource-x509-id and resource-x509-issuer */ 
    if (setResourceX509ID)
    {
        FILE *              lcmaps_pem_fh = NULL;
        X509 *              px509 = NULL;
        #define             PEM_BUF_SIZE 10000
        char                pem_buf[PEM_BUF_SIZE];
        int                 nbytes = 0;

        bzero (pem_buf, PEM_BUF_SIZE);

        /* Read PEM string */
        if ( (lcmaps_pem_fh = fopen(certFilePub, "r")) == NULL)
        {
            lcmaps_log (2, "Could not open file for reading: %s\n", certFilePub);
        }
        else
        {
            nbytes = fread(pem_buf, sizeof(pem_buf), 1, lcmaps_pem_fh);
            if (feof(lcmaps_pem_fh))
            {
                lcmaps_log_debug (3, "%s: EOF reached: %s\n", logstr, certFilePub);
            }

            lcmaps_log_debug (3, "%s: nbytes = %d. Error = %d\n", logstr, nbytes, ferror(lcmaps_pem_fh));
            fclose(lcmaps_pem_fh);

            if (pem_string_to_x509(&px509, pem_buf) != 0)
            {
                lcmaps_log (3, "%s: Error in lcmaps_pem_string_to_x509()\n", logstr);
            }
            else
            {
                resource_x509_id     = X509_NAME_oneline(X509_get_subject_name(px509), NULL, 0);
                resource_x509_issuer = X509_NAME_oneline(X509_get_issuer_name(px509), NULL, 0);

                /* Mandatory clean up - Fix by Brian */
                X509_free(px509);
                px509 = NULL;
            }
        }

    }


    /* Construct resource */
    if (pep_construct_request_resource (request,
                                        resourcetype,
                                        local_host,
                                        NULL,
                                        resource_x509_id,
                                        resource_x509_issuer
                                       ))
    {
        lcmaps_log (0, "%s: XACML: Error in constructing request resource\n", logstr);
        free (local_host); 
        local_host = NULL;
        goto fail_scas_client;
    }

    if (local_host)
    {
        free (local_host);
        local_host = NULL;
    }
    if (resource_x509_id)
    {
        free (resource_x509_id);
        resource_x509_id = NULL;
    }
    if (resource_x509_issuer)
    {
        free (resource_x509_issuer);
        resource_x509_issuer = NULL;
    }
    /************ E: Construction of the XACML request ***********/


    lcmaps_log_debug (5, "%s: XACML framework: Registering Obligation Handlers\n", logstr);

    /************ B: Register Obligation Handlers *********/
    if (pep_register_obligation_handlers (request))
    {
        lcmaps_log (1, "%s: XACML: Error in Registering Obligation Handlers\n", logstr);
        goto fail_scas_client;
    }
    /************ E: Register Obligation Handlers *********/


    if (certFilePub == NULL || certFilePriv == NULL)
    {
        lcmaps_log (1, "%s: XACML: Error can't work without a certificate for the connection to the service\n", logstr);
        goto fail_scas_client;
    }


    /************* B: Silence the stdout and stderr of the SAML2-XACML2 library *************/
    fd_out      = fileno(stdout);
    fd_out_save = dup (fd_out);
    fd_err      = fileno(stderr);
    fd_err_save = dup (fd_err);

    streams_redirected = 1;

    freopen ("/dev/null", "w", stdout);
    freopen ("/dev/null", "w", stderr);
    /************* E: Silence the stdout and stderr of the SAML2-XACML2 library *************/



    /* Execute endpoint selection strategy */
    switch (endpoint_selection_strategy)
    {
        /* Round Robin - Top to Bottom */
        case ENDPOINT_SELECTION_ROUND_ROBIN_TOP_TO_BOTTOM :
            lcmaps_log_debug (5, "%s: Starting Round-Robin (top to bottom) strategy\n", logstr);
            should_do_high_level_retry = 1;

            for (i = 0; i < n_endpoints && should_do_high_level_retry; i++)
            {
                int j = 0;

                for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++)
                {
                    /* Interact with the SCAS */
                    rc = xacmlqueryscas (endpoints[i], request, response);
                    switch (rc)
                    {
                        case XACML_QUERY_ALLOWED        : 
                            lcmaps_log (1, "%s: User \"%s\" with first FQAN \"%s\" Permitted at endpoint \"%s\"\n", 
                                           logstr, user_dn, (nfqans > 0) ? fqans[0] : "", endpoints[i]);
                            should_do_high_level_retry = 0;
                            break;
                        case XACML_QUERY_BANNED         : 
                            lcmaps_log (1, "%s: User \"%s\" with first FQAN \"%s\" Banned at endpoint \"%s\"\n", 
                                           logstr, user_dn, (nfqans > 0) ? fqans[0] : "", endpoints[i]);
                            should_do_high_level_retry = 0;
                            goto fail_scas_client;
                            break;
                        case XACML_QUERY_PROTOCOL_ERROR : 
                            lcmaps_log_debug (2, "%s: Failed to authorize and get a mapping with endpoint \"%s\" (on try #%d)\n", logstr, endpoints[i], j + 1);

                            /* Wait for 50 milli seconds, plus a random number that randomizes below the 10 milli seconds */
                            usleep (50000 + rand_r (&seed) % 10000);
                            break; /* continue the for loops */
                        default:
                            lcmaps_log (1, "%s: Failed to authorize and get a mapping with endpoint (with odd error return from xacmlqueryscas()) \"%s\" (on try #%d)\n", 
                                            logstr, endpoints[i], j + 1);

                            /* Wait for 50 milli seconds, plus a random number that randomizes below the 10 milli seconds */
                            usleep (50000 + rand_r (&seed) % 10000);
                            break; /* continue the for loops */
                    }
                }
            }

            /* All else is an error, besides the XACML_QUERY_ALLOWD */
            if (rc != XACML_QUERY_ALLOWED)
                goto fail_scas_client;

            break;
        case ENDPOINT_SELECTION_ROUND_ROBIN_RANDOM_START :
            lcmaps_log_debug (5, "%s: Starting Round-Robin (random start) strategy\n", logstr);
            should_do_high_level_retry = 1;

            /* Pick a random starting point */
            random_choice = rand_r (&seed) % n_endpoints;

            for (i = 0; i < n_endpoints; i++)
            {
                int selected_endpoint = (random_choice + i) % n_endpoints;
                int j = 0;

                for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++)
                {
                    /* Interact with the SCAS */
                    rc = xacmlqueryscas (endpoints[selected_endpoint], request, response);
                    switch (rc)
                    {
                        case XACML_QUERY_ALLOWED        : 
                            lcmaps_log (1, "%s: User \"%s\" with first FQAN \"%s\" Permitted at endpoint \"%s\"\n", 
                                           logstr, user_dn, (nfqans > 0) ? fqans[0] : "", endpoints[selected_endpoint]);
                            should_do_high_level_retry = 0;
                            break;
                        case XACML_QUERY_BANNED         : 
                            lcmaps_log (1, "%s: User \"%s\" with first FQAN \"%s\" Banned at endpoint \"%s\"\n", 
                                           logstr, user_dn, (nfqans > 0) ? fqans[0] : "", endpoints[selected_endpoint]);
                            should_do_high_level_retry = 0;
                            goto fail_scas_client;
                            break;
                        case XACML_QUERY_PROTOCOL_ERROR : 
                            lcmaps_log_debug (2, "%s: Failed to authorize and get a mapping with endpoint \"%s\" (on try #%d)\n", logstr, endpoints[selected_endpoint], j + 1);

                            /* Wait for 50 milli seconds, plus a random number that randomizes below the 10 milli seconds */
                            usleep (50000 + rand_r (&seed) % 10000);
                            break; /* continue the for loops */
                        default:
                            lcmaps_log (2, "%s: Failed to authorize and get a mapping with endpoint (with odd error return from xacmlqueryscas()) \"%s\" (on try #%d)\n", 
                                            logstr, endpoints[selected_endpoint], j + 1);

                            /* Wait for 50 milli seconds, plus a random number that randomizes below the 10 milli seconds */
                            usleep (50000 + rand_r (&seed) % 10000);
                            break; /* continue the for loops */
                    }
                }
            }

            /* All else is an error, besides the XACML_QUERY_ALLOWD */
            if (rc != XACML_QUERY_ALLOWED)
                goto fail_scas_client;

            break;
        case ENDPOINT_SELECTION_RANDOM :
            lcmaps_log_debug (5, "%s: Starting Random selection strategy\n", logstr);
            should_do_high_level_retry = 1;
            
            for (i = 0; i < n_endpoints; i++)
            {
                /* Pick a random choice */
                random_choice = rand_r (&seed) % n_endpoints;

                int j = 0;

                for (j = 0; j < endpoint_retrycount && should_do_high_level_retry; j++)
                {
                    /* Interact with the SCAS */
                    rc = xacmlqueryscas (endpoints[random_choice], request, response);
                    switch (rc)
                    {
                        case XACML_QUERY_ALLOWED        : 
                            lcmaps_log (1, "%s: User \"%s\" with first FQAN \"%s\" Permitted at endpoint \"%s\"\n", 
                                           logstr, user_dn, (nfqans > 0) ? fqans[0] : "", endpoints[random_choice]);
                            should_do_high_level_retry = 0;
                            break;
                        case XACML_QUERY_BANNED         : 
                            lcmaps_log (1, "%s: User \"%s\" with first FQAN \"%s\" Banned at endpoint \"%s\"\n", 
                                           logstr, user_dn, (nfqans > 0) ? fqans[0] : "", endpoints[random_choice]);
                            should_do_high_level_retry = 0;
                            goto fail_scas_client;
                            break;
                        case XACML_QUERY_PROTOCOL_ERROR : 
                            lcmaps_log_debug (2, "%s: Failed to authorize and get a mapping with endpoint \"%s\" (on try #%d)\n", logstr, endpoints[random_choice], j + 1);

                            /* Wait for 50 milli seconds, plus a random number that randomizes below the 10 milli seconds */
                            usleep (50000 + rand_r (&seed) % 10000);
                            break; /* continue the for loops */
                        default:
                            lcmaps_log (2, "%s: Failed to authorize and get a mapping with endpoint (with odd error return from xacmlqueryscas()) \"%s\" (on try #%d)\n", 
                                            logstr, endpoints[random_choice], j + 1);

                            /* Wait for 50 milli seconds, plus a random number that randomizes below the 10 milli seconds */
                            usleep (50000 + rand_r (&seed) % 10000);
                            break; /* continue the for loops */
                    }
                }
            }

            /* All else is an error, besides the XACML_QUERY_ALLOWD */
            if (rc != XACML_QUERY_ALLOWED)
                goto fail_scas_client;

            break;
        default : 
            /* Default to Round Robin - Top to Bottom */
            lcmaps_log_debug (5, "%s: Starting Round-Robin (top to bottom) strategy (default)\n", logstr);
            should_do_high_level_retry = 1;

            for (i = 0; i < n_endpoints; i++)
            {
                int j = 0;

                for (j = 0; (j < endpoint_retrycount) && should_do_high_level_retry; j++)
                {
                    /* Interact with the SCAS */
                    rc = xacmlqueryscas (endpoints[i], request, response);
                    switch (rc)
                    {
                        case XACML_QUERY_ALLOWED        : 
                            lcmaps_log (1, "%s: User \"%s\" with first FQAN \"%s\" Permitted at endpoint \"%s\"\n", 
                                           logstr, user_dn, (nfqans > 0) ? fqans[0] : "", endpoints[i]);
                            should_do_high_level_retry = 0;
                            break;
                        case XACML_QUERY_BANNED         : 
                            lcmaps_log (1, "%s: User \"%s\" with first FQAN \"%s\" Banned at endpoint \"%s\"\n", 
                                           logstr, user_dn, (nfqans > 0) ? fqans[0] : "", endpoints[i]);
                            should_do_high_level_retry = 0;
                            goto fail_scas_client;
                            break;
                        case XACML_QUERY_PROTOCOL_ERROR : 
                            lcmaps_log_debug (2, "%s: Failed to authorize and get a mapping with endpoint \"%s\" (on try #%d)\n", logstr, endpoints[i], j + 1);

                            /* Wait for 50 milli seconds, plus a random number that randomizes below the 10 milli seconds */
                            usleep (50000 + rand_r (&seed) % 10000);
                            break; /* continue the for loops */
                        default:
                            lcmaps_log (2, "%s: Failed to authorize and get a mapping with endpoint (with odd error return from xacmlqueryscas()) \"%s\" (on try #%d)\n", logstr, endpoints[i], j + 1);

                            /* Wait for 50 milli seconds, plus a random number that randomizes below the 10 milli seconds */
                            usleep (50000 + rand_r (&seed) % 10000);
                            break; /* continue the for loops */
                    }
                }
            }

            /* All else is an error, besides the XACML_QUERY_ALLOWD */
            if (rc != XACML_QUERY_ALLOWED)
                goto fail_scas_client;

            break;
    }


    /* This fix is to adjust for LCMAPS version 1.4.6 and older */
    if (use_pool_index_bug_fix)
    {
        lcmaps_log_debug (5, "%s: Adding POOL_INDEX :  %s\n", logstr, encoded_dn);
        addCredentialData(POOL_INDEX, &encoded_dn);
    }



/* Succesful clean up */
    lcmaps_log_debug (5, "%s: Destroying: Obligation handlers\n", logstr);

/*
    xacml_obligation_destroy(UIDGID_handler); 
    xacml_obligation_destroy(SecondaryGIDs_handler); 
*/

/*
 default_handler
 local_user_name_handler
 UIDGID_handler
 SecondaryGIDs_handler
 AFSToken_handler
*/

    lcmaps_log_debug (5, "%s: Destroying: response\n", logstr);
    xacml_response_destroy(response); response=NULL; /* set to NULL to prevent double free */
    lcmaps_log_debug (5, "%s: Destroying: request\n", logstr);
    xacml_request_destroy(request); request=NULL; /* set to NULL to prevent double free */

    ///////////////////////////
    ///////////////////////////


    /* Cleanup */

    /** B: Restore stderr and stdout streams ***/
    if (streams_redirected)
    {
        fflush (stdout);
        fflush (stderr);
        dup2 (fd_out_save, fd_out);
        close(fd_out_save);
        dup2 (fd_err_save, fd_err);
        close(fd_err_save);

        streams_redirected = 0;
    }
    /** B: Restore stderr and stdout streams ***/


    /* Free stuff */
    if (datetime)             { free(datetime); datetime = NULL; }
    if (user_dn)	      { free(user_dn); user_dn = NULL; }
    if (issuer_dn)	      { free(issuer_dn); issuer_dn=NULL; }
    if (fqans)		      { free(fqans); fqans=NULL; }


    if (local_host)           { free (local_host); local_host = NULL; }
    if (resource_x509_id)     { free (resource_x509_id); resource_x509_id = NULL; }
    if (resource_x509_issuer) { free (resource_x509_issuer); resource_x509_issuer = NULL; }

    lcmaps_log_debug (3,"%s: scas client plugin succeeded\n", logstr);

    return LCMAPS_MOD_SUCCESS;


fail_scas_client:

    /** B: Restore stderr and stdout streams ***/
    if (streams_redirected)
    {
        fflush (stdout);
        fflush (stderr);
        dup2 (fd_out_save, fd_out);
        close(fd_out_save);
        dup2 (fd_err_save, fd_err);
        close(fd_err_save);

        streams_redirected = 0;
    }
    /** B: Restore stderr and stdout streams ***/

    /* Free stuff */
    if (user_dn)	       { free( user_dn ); user_dn=NULL; }
    if (issuer_dn)	       { free( issuer_dn ); issuer_dn=NULL; }
    if (fqans)		       { free( fqans ); fqans=NULL; }

    lcmaps_log_debug (5, "%s: Destroying: Obligation handlers\n", logstr);
    lcmaps_log_debug (5, "%s: Destroying: response\n", logstr);
    xacml_response_destroy(response); response=NULL; /* set to NULL to prevent double free */
    lcmaps_log_debug (5, "%s: Destroying: request\n", logstr);
    xacml_request_destroy(request); request=NULL; /* set to NULL to prevent double free */

    if (datetime)             { free(datetime); datetime = NULL; }

    if (local_host)           { free (local_host); local_host = NULL; }
    if (resource_x509_id)     { free (resource_x509_id); resource_x509_id = NULL; }
    if (resource_x509_issuer) { free (resource_x509_issuer); resource_x509_issuer = NULL; }

    lcmaps_log_time(2,"%s: scas client plugin failed\n", logstr);

    return LCMAPS_MOD_FAIL;
}



/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCMAPS_MOD_SUCCESS : succes
    LCMAPS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate(void)
{
    char * logstr = "\tlcmaps_plugin_scas_client-plugin_terminate()";
    lcmaps_log_debug(1,"%s: terminating\n", logstr);

    if (certFilePub)          { free(certFilePub);  certFilePub  = NULL; }
    if (certFilePriv)         { free(certFilePriv); certFilePriv = NULL; }
    if (certCAPath)           { free(certCAPath);   certCAPath   = NULL; }
    if (certOwner)            { free(certOwner);    certOwner    = NULL; }

    return LCMAPS_MOD_SUCCESS;
}



/******************************************************************************
Function:    collapseFqansFromVOMSData
Description:
             Create an array of FQANs from the VOMS Data structure
             As it is implemented, I'm only referencing pointers
             Don't free the values, only the returned char ** fqans.
Parameters:
             lcmaps_vomsdata_t *   vomsdata : LCMAPS replication of the vomsdata structure (and something extra)
             char              *** fqans    : the returned FQANs which is to be the char ** array. 
                                              For he who would understand, yes... I'm using a triple pointer...
             int               *            : amount of returned FQANs in the array
Returns:
    0 : good
   !0 : bad
******************************************************************************/

int collapseFqansFromVOMSData (lcmaps_vomsdata_t *   vomsdata,
                               char              *** fqans,
                               int               *   nfqans)
{
    int i   = 0;
    int j   = 0;
    int k   = 0;

    char ** f = NULL;
    int     nf = 0;

/*
    *fqans  = NULL;
    *nfqans = 0;
*/

    /* Count all the FQANs from multiple vomsdata structs */
    for (i = 0; i < vomsdata->nvoms; i++)
    {
       nf += vomsdata->voms[i].nfqan;
    }

    /* Allocate the FQAN slots */
    f = (char **) malloc (nf * sizeof (char *));
    for (i = 0; i < vomsdata->nvoms; i++)
    {
        for (j = 0; j < vomsdata->voms[i].nfqan; j++)
        {
            f[k++] = vomsdata->voms[i].fqan_unix[j].fqan;
        }
    }

    *nfqans = nf;
    *fqans  = f;

    return 0;
}


/******************************************************************************
Function:       pem_string_to_x509
Description:    Reads a X509 certificate from a PEM-encoded string
Parameters:
                px: pointer to a X509 certificate
                cert_string: PEM-encoded string 
Returns:        0 on success, non-zero otherwise
******************************************************************************/
int pem_string_to_x509(X509 **px, char *certstring)
{

    BIO *certbio = NULL;

    *px = X509_new();
    if (*px == NULL) return -1;
    /* clean it first */
    X509_free(*px);

    certbio = BIO_new_mem_buf(certstring, -1);
    if (certbio)
    {
        /* This call allocates new X509 struct, so no X509_new() needed */
        *px = PEM_read_bio_X509(certbio, NULL, 0, NULL);
    }

    BIO_free(certbio);
    return 0;
}


/******************************************************************************
Function:       xacmlqueryscas
Description:    Queries the SCAS service for a authorization decision and mapping
Parameters:
                endpoint
                request
                response
Returns:        0 on success, non-zero otherwise
******************************************************************************/
XACML_QUERY_ANSWER_T xacmlqueryscas (char *           endpoint, 
                                     xacml_request_t  request, 
                                     xacml_response_t response)
{
    char * logstr = "\txacmlqueryscas()";
    saml_status_code_t   code;
    xacml_decision_t     decision;
    int                  rc_xacml = 0;

    lcmaps_log_debug (2, "%s: XACML: Using SCAS service endpoint: %s\n", logstr, endpoint);


    /* Query XACML endpoint */
    rc_xacml = xacml_query(endpoint, request, response);

    switch (rc_xacml)
    {
        case XACML_RESULT_SUCCESS :
            lcmaps_log_debug (3, "%s: XACML: Success with endpoint: %s\n", logstr, endpoint);

            xacml_response_get_saml_status_code(response, &code);
            xacml_response_get_xacml_decision(response, &decision);
                    
            lcmaps_log_debug (5, "%s: Server \"%s\" said: \"%s\" with decision: \"%s\"\n", 
                                logstr, 
                                endpoint, 
                                saml_status_code_strings[code], 
                                decision == XACML_DECISION_Permit ? "Permit" : 
                                    (decision == XACML_DECISION_Deny ? "Deny" : 
                                        (decision == XACML_DECISION_Indeterminate ? "Indeterminate" : 
                                            (decision == XACML_DECISION_NotApplicable ? "NotApplicable" : "Unknown"))));

            switch (decision)
            {
                case XACML_DECISION_Permit        :
                    /* Specific fix to be compatible with service like SAZ that do not give out obligations - this must be expplicitly activated */
                    if (use_authorization_only)
                    {
                        lcmaps_log_debug (5, "%s: XACML: Authorized by service \"%s\"\n", logstr, endpoint);
                        return XACML_QUERY_ALLOWED;
                    }
                    else
                    {
                        /* 
                         * This patch was supplied by Brian Bockelman 
                         * - it fixes the bug that we should check both the Username Obligation Handler or the UIDGID OH to be fired.
                         *   This happens when the GUMS service is used.
                         */
                        if (getOHUIDGIDFired())
                        {
                            lcmaps_log_debug (5, "%s: XACML: Got a UIDGID Obligation (meaning: sufficient Unix account mapping information) by \"%s\"\n", logstr, endpoint);
                            return XACML_QUERY_ALLOWED;
                        }
                        else if (getOHUsernameFired())
                        {
                            lcmaps_log_debug (5, "%s: XACML: Got a Username Obligation (meaning: sufficient Unix account mapping information) by \"%s\"\n", logstr, endpoint);
                            return XACML_QUERY_ALLOWED;
                        }
                        else
                        {
                            lcmaps_log (1, "%s: XACML: No mapping result. This is an unknown state. Endpoint used: \"%s\"\n", logstr, endpoint);
                            return XACML_QUERY_PROTOCOL_ERROR;
                        }
                    }
                case XACML_DECISION_Deny          :
                    lcmaps_log (1, "%s: XACML: Access denied at endpoint: \"%s\"\n", logstr, endpoint);
                    return XACML_QUERY_BANNED;
                case XACML_DECISION_Indeterminate :
                    lcmaps_log (1, "%s: XACML: Failure: endpoint returned \"Indeterminate\". This is an unsupportable state by \"%s\"\n", logstr, endpoint);
                    return XACML_QUERY_PROTOCOL_ERROR;
                case XACML_DECISION_NotApplicable :
                    lcmaps_log (1, "%s: XACML: Failure: endpoint returned \"Not Applicable\". This is an unsupportable state by \"%s\"\n", logstr, endpoint);
                    return XACML_QUERY_PROTOCOL_ERROR;
            }        
            break;
        case XACML_RESULT_INVALID_PARAMETER :
            lcmaps_log (1, "%s: XACML: Interaction failed: Invalid parameter. Endpoint is: \"%s\"\n", logstr, endpoint);
            return XACML_QUERY_PROTOCOL_ERROR;
        case XACML_RESULT_OBLIGATION_FAILED :
            lcmaps_log (1, "%s: XACML: Interaction failed: Obligation failed. The returned obligations couldn't be processed from endpoint: \"%s\"\n", logstr, endpoint);
            return XACML_QUERY_PROTOCOL_ERROR;
        case XACML_RESULT_SOAP_ERROR :
            lcmaps_log (1, "%s: XACML: Interaction failed: TCP/IP, SSL or SOAP Error with endpoint: \"%s\"\n", logstr, endpoint);
            return XACML_QUERY_PROTOCOL_ERROR;
        case XACML_RESULT_INVALID_STATE :
            lcmaps_log (1, "%s: XACML: Interaction failed: Invalid state. Endpoint is: \"%s\"\n", logstr, endpoint);
            return XACML_QUERY_PROTOCOL_ERROR;
        default :
            lcmaps_log (1, "%s: XACML: Interaction failed: Unknown error. Endpoint is: \"%s\"\n", logstr, endpoint);
            return XACML_QUERY_PROTOCOL_ERROR;
    }

    return XACML_QUERY_ALLOWED;
}
